from typing import List

import ifcopenshell
import numpy as np
from compas.tolerance import TOL

from .primities import frame_to_ifc_plane
from .primities import occ_plane_to_frame
from .primities import point_to_ifc_cartesian_point
from .shapes import occ_cylinder_to_ifc_cylindrical_surface


def calculate_knots_and_multiplicities(knot_sequence):
    knots = [knot_sequence[0]]
    multiplicities = [1]

    for i in range(1, len(knot_sequence)):
        if knot_sequence[i] != knot_sequence[i - 1]:
            knots.append(knot_sequence[i])
            multiplicities.append(1)
        else:
            multiplicities[-1] += 1

    return knots, multiplicities


def brep_to_ifc_advanced_brep(file: ifcopenshell.file, brep) -> List[ifcopenshell.entity_instance]:
    brep.fix()
    brep.sew()
    brep.make_solid()

    points = {}
    curves = {}
    lines = {}

    def get_ifc_point(point):
        key = TOL.geometric_key(point)
        if key in points:
            return points[key]
        points[key] = point_to_ifc_cartesian_point(file, point)
        return points[key]

    def get_ifc_line(edge):
        line_key = TOL.geometric_key(edge.first_vertex.point) + "-" + TOL.geometric_key(edge.last_vertex.point)
        return lines.get(line_key)

    def get_ifc_curve(edge):
        curve = edge.nurbscurve
        for occ_curve in curves:
            if occ_curve.IsEqual(curve.occ_curve, 1e-6):
                return curves[occ_curve]

    ifc_breps = []

    for edge in brep.edges:
        if edge.is_bspline:
            if get_ifc_curve(edge):
                continue

            start_vertex = file.create_entity("IfcVertexPoint", get_ifc_point(edge.first_vertex.point))
            if edge.nurbscurve.is_closed:
                end_vertex = start_vertex
            else:
                end_vertex = file.create_entity("IfcVertexPoint", get_ifc_point(edge.last_vertex.point))

            curve = edge.nurbscurve
            control_points = [get_ifc_point(point) for point in curve.points]
            weights = curve.weights

            # OCC will simplify the knot and multiplicity when the curver or surface is periodic,
            # so we need to recalculate the knot and multiplicity from the knot sequence.
            # Additonally, we will need to add back the duplicated control point and weight value.
            knots, multiplicities = calculate_knots_and_multiplicities(curve.knotsequence)
            if curve.is_closed:
                control_points.append(control_points[0])
                weights.append(weights[0])

            IfcBSpline = file.create_entity(
                "IfcRationalBSplineCurveWithKnots",
                Degree=curve.degree,
                ControlPointsList=control_points,
                CurveForm="UNSPECIFIED",
                ClosedCurve=curve.is_closed,
                SelfIntersect=False,
                KnotMultiplicities=multiplicities,
                Knots=knots,
                WeightsData=weights,
            )

            IfcEdgeCurve = file.create_entity(
                "IfcEdgeCurve",
                EdgeStart=start_vertex,
                EdgeEnd=end_vertex,
                EdgeGeometry=IfcBSpline,
                SameSense=True,
            )

            curves[curve.occ_curve] = IfcEdgeCurve

        elif edge.is_line:
            if get_ifc_line(edge):
                continue

            start_point = get_ifc_point(edge.first_vertex.point)
            end_point = get_ifc_point(edge.last_vertex.point)
            start_vertex = file.create_entity("IfcVertexPoint", get_ifc_point(edge.first_vertex.point))
            end_vertex = file.create_entity("IfcVertexPoint", get_ifc_point(edge.last_vertex.point))

            IfcPolyLine = file.create_entity(
                "IfcPolyLine",
                Points=[start_point, end_point],
            )

            IfcEdgeCurve = file.create_entity(
                "IfcEdgeCurve",
                EdgeStart=start_vertex,
                EdgeEnd=end_vertex,
                EdgeGeometry=IfcPolyLine,
                SameSense=True,
            )

            line_key = TOL.geometric_key(edge.first_vertex.point) + "-" + TOL.geometric_key(edge.last_vertex.point)
            lines[line_key] = IfcEdgeCurve

        else:
            raise NotImplementedError("Only BSpline and Line edges are supported")

    for solid in brep.solids:
        for shell in solid.shells:
            ifc_faces = []
            for face in shell.faces:
                face_bounds = []
                is_outer = True

                for loop in face.loops:
                    ifc_oriented_edges = []

                    for edge in loop.edges:
                        if edge.is_bspline:
                            oriented = edge.occ_edge.Orientation() == 0
                            IfcEdgeCurve = get_ifc_curve(edge)
                        elif edge.is_line:
                            oriented = edge.occ_edge.Orientation() == 0
                            IfcEdgeCurve = get_ifc_line(edge)
                        else:
                            raise NotImplementedError("Only BSpline and Line edges are supported")

                        if not IfcEdgeCurve:
                            raise ValueError("Edge not found")

                        ifc_oriented_edge = file.create_entity("IFCORIENTEDEDGE", EdgeElement=IfcEdgeCurve, Orientation=oriented)
                        ifc_oriented_edges.append(ifc_oriented_edge)

                    edge_loop = file.create_entity("IfcEdgeLoop", ifc_oriented_edges)
                    if is_outer:
                        ifc_face_bound = file.create_entity("IfcFaceOuterBound", edge_loop, True)
                        is_outer = False
                    else:
                        ifc_face_bound = file.create_entity("IfcFaceBound", edge_loop, True)
                    face_bounds.append(ifc_face_bound)

                same_sense = face.orientation == 0
                if face.is_plane:
                    occ_plane = face.occ_adaptor.Plane()
                    frame = occ_plane_to_frame(occ_plane)
                    ifc_plane = frame_to_ifc_plane(file, frame)
                    IfcAdvancedFace = file.create_entity("IfcAdvancedFace", face_bounds, ifc_plane, SameSense=same_sense)
                elif face.is_cylinder:
                    cylinder = face.occ_adaptor.Cylinder()
                    IfcCylindricalSurface = occ_cylinder_to_ifc_cylindrical_surface(file, cylinder)
                    IfcAdvancedFace = file.create_entity("IfcAdvancedFace", face_bounds, IfcCylindricalSurface, SameSense=same_sense)
                else:
                    control_points = np.array(face.nurbssurface.points.points, dtype=float)
                    control_points = control_points.swapaxes(0, 1)
                    ifc_control_points = []

                    u_knots, u_mults = calculate_knots_and_multiplicities(list(face.nurbssurface.occ_surface.UKnotSequence()))
                    v_knots, v_mults = calculate_knots_and_multiplicities(list(face.nurbssurface.occ_surface.VKnotSequence()))

                    for row in control_points:
                        ifc_row = []
                        for point in row:
                            ifc_row.append(get_ifc_point(point))
                        ifc_control_points.append(ifc_row)

                    if face.nurbssurface.is_periodic_u:
                        new_row = []
                        for point in control_points[0]:
                            new_row.append(get_ifc_point(point))
                        ifc_control_points.append(new_row)

                    if face.nurbssurface.is_periodic_v:
                        for i, row in enumerate(ifc_control_points):
                            row.append(get_ifc_point(control_points[i % len(control_points)][0]))

                    weights = face.nurbssurface.weights
                    weights = np.array(weights, dtype=float)
                    weights = weights.swapaxes(0, 1)

                    ifc_weights = []
                    for row in weights:
                        ifc_row = []
                        for weight in row:
                            ifc_row.append(float(weight))
                        ifc_weights.append(ifc_row)

                    if face.nurbssurface.is_periodic_u:
                        ifc_weights.append(ifc_weights[0])
                    if face.nurbssurface.is_periodic_v:
                        for i, row in enumerate(ifc_weights):
                            row.append(row[0])

                    IfcBSplineSurfaceWithKnots = file.create_entity(
                        "IfcRationalBSplineSurfaceWithKnots",
                        UDegree=face.nurbssurface.degree_u,
                        VDegree=face.nurbssurface.degree_v,
                        ControlPointsList=ifc_control_points,
                        SurfaceForm="UNSPECIFIED",
                        UClosed=face.nurbssurface.is_periodic_u,
                        VClosed=face.nurbssurface.is_periodic_v,
                        SelfIntersect=False,  # Seems no way to get this from OCC
                        UMultiplicities=u_mults,
                        VMultiplicities=v_mults,
                        UKnots=u_knots,
                        VKnots=v_knots,
                        WeightsData=ifc_weights,
                    )

                    IfcAdvancedFace = file.create_entity("IfcAdvancedFace", face_bounds, IfcBSplineSurfaceWithKnots, SameSense=same_sense)

                ifc_faces.append(IfcAdvancedFace)

            ifc_shell = file.create_entity("IFCCLOSEDSHELL", ifc_faces)

        ifc_brep = file.create_entity("IFCADVANCEDBREP", ifc_shell)
        ifc_breps.append(ifc_brep)

    return ifc_breps



from typing import TYPE_CHECKING

import ifcopenshell.guid
from ifcopenshell.api import run
from ifcopenshell.util.element import get_psets

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcObject
else:
    IfcObject = object


class IfcObject(IfcObject):
    _psetsmap = {}

    @property
    def properties(self):
        psets = get_psets(self.entity, psets_only=True)
        for pset in psets.values():
            del pset["id"]
        return psets

    @properties.setter
    def properties(self, psets):
        if id(self.file) not in self._psetsmap:
            self._psetsmap[id(self.file)] = {}

        psetsmap = self._psetsmap[id(self.file)]

        for name, properties in psets.items():
            if id(properties) in psetsmap:
                pset = psetsmap[id(properties)]
                # TODO: Check if relation already exists
                self.file._create_entity(
                    "IfcRelDefinesByProperties",
                    GlobalId=ifcopenshell.guid.new(),
                    OwnerHistory=self.file.default_owner_history,
                    RelatingPropertyDefinition=pset,
                    RelatedObjects=[self.entity],
                )

            else:
                pset = run("pset.add_pset", self.file._file, product=self.entity, name=name)
                psetsmap[id(properties)] = pset
                for key, value in properties.items():
                    if not isinstance(value, (str, int, float)):
                        properties[key] = str(value)
                run("pset.edit_pset", self.file._file, pset=pset, properties=properties)
                # TODO: remove unused psets

    @property
    def quantities(self):
        qtos = get_psets(self.entity, qtos_only=True)
        for qto in qtos.values():
            del qto["id"]
        return qtos



from typing import List

from compas_ifc.entities.product import Product


class SpatialElement(Product):
    """
    Class representing an IFC spatial element. A spatial element is a product that is used to spatially organize other products.
    """

    def is_containment_hierarchical(self) -> bool:
        """Return True if the spatial element is hierarchical, which means it can have only one decomposition parent."""
        if len(self._entity.Decomposes) > 1:
            return False
        return True

    def contains_elements(self) -> List["SpatialElement"]:
        """Returen contained elements of a spatial element."""
        return [self.model.reader.get_entity(rel) for rel in self._entity.ContainsElements]



from .product import Product


class Element(Product):
    """
    Class representing an IFC element. An element is a product that is intended to be physically constructed or installed.

    Attributes
    ----------
    parent : :class:`compas_ifc.entities.SpatialElement`
        The spatial element containing this element.

    """

    def contained_in_structure(self):
        """Return the spatial structure containing this element."""
        if self not in self.model._new_entities:
            for rel in self._entity.ContainedInStructure:
                return self.model.reader.get_entity(rel)

    @property
    def parent(self):
        if not self._parent and self._entity:
            relation = self.contained_in_structure()
            if relation:
                self._parent = relation["RelatingStructure"]
            else:
                relation = self.decomposes()
                if relation:
                    self._parent = relation["RelatingObject"]
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent



import multiprocessing
import os
import time
from typing import Any
from typing import Union

import ifcopenshell
import numpy as np
from compas.geometry import Transformation
from ifcopenshell.api import run

import compas_ifc
from compas_ifc.entities.base import Base


class IFCFile(object):
    def __init__(self, model, filepath=None, schema="IFC4", use_occ=False, load_geometries=True, verbose=True):
        self.verbose = verbose
        self.ensure_classes_generated()
        self._entitymap = {}
        self._geometrymap = {}
        self._stylemap = {}
        self._relationmap_aggregates = {}  # map of IfcRelAggregates
        self._relationmap_contains = {}  # map of IfcRelContainedInSpatialStructure
        self._default_context = None
        self._default_body_context = None
        self._default_units = None
        self._default_owner_history = None
        self._default_project = None

        self.filepath = filepath
        self.model = model
        self.use_occ = use_occ
        if filepath is None:
            self._file = ifcopenshell.file(schema=schema)
            if self.verbose:
                print("IFC file created in schema: {}".format(schema))
        else:
            self._file = ifcopenshell.open(filepath)
            if self.verbose:
                print("IFC file loaded: {}".format(filepath))

        self._schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(self._file.schema)

        if load_geometries and filepath is not None:
            self.load_geometries()

    @property
    def schema(self):
        return self._schema

    @property
    def schema_name(self):
        return self.schema.name()

    def ensure_classes_generated(self):
        try:
            from compas_ifc.entities.generated import IFC2X3  # noqa: F401
            from compas_ifc.entities.generated import IFC4  # noqa: F401
        except ImportError:
            if self.verbose:
                print("IFC classes not found. Generating classes...")
            from compas_ifc.entities.generator import Generator

            generator = Generator(schema="IFC2X3")
            generator.generate()

            generator = Generator(schema="IFC4")
            generator.generate()
            if self.verbose:
                print("IFC classes generated.\n\n")

    def file_size(self):
        file_stats = os.stat(self.filepath)
        size_in_mb = file_stats.st_size / (1024 * 1024)
        size_in_mb = round(size_in_mb, 2)
        return size_in_mb

    def from_entity(self, entity):
        if not isinstance(entity, ifcopenshell.entity_instance):
            raise TypeError("Input is not an ifcopenshell.entity_instance")

        _id = entity.id()

        if _id in self._entitymap and _id != 0:
            return self._entitymap[_id]
        else:
            entity = Base(entity, self)
            if _id != 0:
                self._entitymap[_id] = entity
            return entity

    def get_entities_by_type(self, type_name) -> list[Base]:
        entities = self._file.by_type(type_name)
        return [self.from_entity(entity) for entity in entities]

    def get_entities_by_name(self, name) -> list[Base]:
        return [entity for entity in self.get_entities_by_type("IfcRoot") if entity.Name == name]

    def get_entity_by_global_id(self, global_id) -> Base:
        return self.from_entity(self._file.by_guid(global_id))

    def get_entity_by_id(self, id) -> Base:
        return self.from_entity(self._file.by_id(id))

    def get_preloaded_geometry(self, entity):
        return self._geometrymap.get(entity.entity.id())

    def get_preloaded_style(self, entity):
        return self._stylemap.get(entity.entity.id(), {})

    def load_geometries(self, include=None, exclude=None):
        """Load all the geometries of the IFC file using a fast multithreaded iterator."""
        if self.verbose:
            print("Loading geometries...")
        import ifcopenshell.geom

        settings = ifcopenshell.geom.settings()
        settings.set(settings.CONVERT_BACK_UNITS, True)
        if self.use_occ:
            settings.set(settings.USE_PYTHON_OPENCASCADE, True)

        iterator = ifcopenshell.geom.iterator(settings, self._file, multiprocessing.cpu_count(), include=include, exclude=exclude)
        start = time.time()
        if iterator.initialize():
            while True:
                shape = iterator.get()
                if self.use_occ:
                    from compas_occ.brep import OCCBrep

                    brep = OCCBrep.from_shape(shape.geometry)

                    shellcolors = []
                    for style_id, style in zip(shape.style_ids, shape.styles):
                        if style_id == -1:
                            shellcolors.append((0.5, 0.5, 0.5, 1.0))
                        else:
                            shellcolors.append(style)

                    self._geometrymap[shape.data.id] = brep
                    self._stylemap[shape.data.id] = {"shellcolors": shellcolors}

                else:
                    from .brep import TessellatedBrep

                    matrix = shape.transformation.matrix.data
                    faces = shape.geometry.faces
                    edges = shape.geometry.edges
                    verts = shape.geometry.verts

                    matrix = np.array(matrix).reshape((4, 3))
                    matrix = np.hstack([matrix, np.array([[0], [0], [0], [1]])])
                    matrix = matrix.transpose()
                    transformation = Transformation.from_matrix(matrix.tolist())

                    facecolors = []
                    for m_id in shape.geometry.material_ids:
                        if m_id == -1:
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            continue
                        material = shape.geometry.materials[m_id]
                        color = (*material.diffuse, 1 - material.transparency)
                        facecolors.append(color)
                        facecolors.append(color)
                        facecolors.append(color)

                    brep = TessellatedBrep(vertices=verts, edges=edges, faces=faces)

                    brep.transform(transformation)
                    self._geometrymap[shape.id] = brep
                    self._stylemap[shape.id] = {"facecolors": facecolors}

                if not iterator.next():
                    break

        if self.verbose:
            print(f"Time to load all {len(self._geometrymap)} geometries {(time.time() - start):.3f}s")

    def save(self, path):
        self._file.write(path)

    def export(self, path: str, entities: list[Base] = [], as_snippet: bool = False, export_materials: bool = True, export_properties: bool = True, export_styles: bool = True):
        new_file = IFCFile(None, schema=self.schema_name)

        exported = {}

        def export_entity(entity: Union[Base, Any], file: IFCFile):
            if not isinstance(entity, Base):
                return entity

            if entity.is_a("IfcOwnerHistory"):
                return file.default_owner_history

            if entity in exported:
                return exported[entity]

            parent = getattr(entity, "parent", None)
            if parent and not as_snippet:
                new_parent_entity = export_entity(parent, file)

            new_entity = file._create_entity(entity.is_a())
            exported[entity] = new_entity

            if parent and not as_snippet:
                if hasattr(entity, "ContainedInStructure") and entity.ContainedInStructure():
                    file._create_entity("IfcRelContainedInSpatialStructure", RelatingStructure=new_parent_entity, RelatedElements=[new_entity])
                else:
                    file._create_entity("IfcRelAggregates", RelatingObject=new_parent_entity, RelatedObjects=[new_entity])

            for key, attr in entity.attributes.items():
                # Skip Representation and ObjectPlacement if not in the list of entities to export
                if key in ["Representation", "ObjectPlacement"] and entity not in entities:
                    continue

                if isinstance(attr, Base):
                    attr = export_entity(attr, file)
                elif isinstance(attr, (list, tuple)):
                    attr = [export_entity(a, file) for a in attr]

                setattr(new_entity, key, attr)

            if export_properties and hasattr(entity, "properties"):
                new_entity.properties = entity.properties

            if export_materials and hasattr(entity, "HasAssociations"):
                # TODO: create material settor on class extension.
                for relation in entity.HasAssociations():
                    if relation.is_a("IfcRelAssociatesMaterial"):
                        new_material_entity = export_entity(relation.RelatingMaterial, file)
                        file._create_entity("IfcRelAssociatesMaterial", RelatingMaterial=new_material_entity, RelatedObjects=[new_entity])

            if export_styles and hasattr(entity, "StyledByItem") and entity.entity.id():
                # TODO: create style settor on class extension.
                for style_item in entity.StyledByItem():
                    styles = [export_entity(s, file) for s in style_item.Styles]
                    file._create_entity("IfcStyledItem", Styles=styles, Item=new_entity)

            return new_entity

        for entity in entities:
            export_entity(entity, new_file)

        new_file.save(path)

    def create(self, cls="IfcBuildingElementProxy", parent=None, geometry=None, frame=None, properties=None, **kwargs):
        if isinstance(cls, type):
            cls_name = cls.__name__
        else:
            cls_name = cls

        entity = self._create_entity(cls_name, **kwargs)

        if parent:
            if hasattr(entity, "ContainedInStructure"):
                self._create_entity("IfcRelContainedInSpatialStructure", RelatingStructure=parent, RelatedElements=[entity])
            else:
                self._create_entity("IfcRelAggregates", RelatingObject=parent, RelatedObjects=[entity])

        if geometry:
            # TODO: Deal with instancing
            entity.geometry = geometry

        if frame:
            entity.frame = frame

        if properties:
            entity.properties = properties

        return entity

    def remove(self, entity):
        if isinstance(entity, Base):
            entity = [entity]

        print(f"Removing {len(entity)} entities...")
        ifcopenshell.util.element.batch_remove_deep2(self._file)
        for e in entity:
            ifcopenshell.util.element.remove_deep2(self._file, e.entity)
        self._file = ifcopenshell.util.element.unbatch_remove_deep2(self._file)
        print("Removal done.")

    def _create_entity(self, cls_name, **kwargs):
        camel_case_kwargs = {}

        for key, value in kwargs.items():
            if isinstance(value, Base):
                kwargs[key] = value.entity
            elif isinstance(value, (list, tuple)):
                kwargs[key] = [v.entity if isinstance(v, Base) else v for v in value]

            # if key is a snake_case key, convert it to camelCase
            if key[0].isupper():
                camel_case_kwargs[key] = kwargs[key]
            else:
                camel_case_key = "".join([word.capitalize() for word in key.split("_")])
                camel_case_kwargs[camel_case_key] = kwargs[key]

        entity = self._file.create_entity(cls_name, **camel_case_kwargs)
        return self.from_entity(entity)

    @property
    def default_project(self):
        projects = self._file.by_type("IfcProject")
        if projects:
            self._default_project = self.from_entity(projects[0])
        else:
            self._default_project = self._create_entity("IfcProject", Name="Default Project")
            self.default_units
            self.default_body_context
            self.default_owner_history
            return self._default_project

        return self._default_project

    @property
    def default_units(self):
        if not self._default_units:
            if self.default_project.UnitsInContext:
                self._default_units = self.default_project.UnitsInContext
            else:
                self._default_units = self.from_entity(run("unit.assign_unit", self._file))
        return self._default_units

    @property
    def default_context(self):
        if not self._default_context:
            contexts = self.get_entities_by_type("IfcGeometricRepresentationContext")
            for context in contexts:
                if context.ContextType == "Model":
                    self._default_context = context
                    return self._default_context

            self._default_context = self.from_entity(run("context.add_context", self._file, context_type="Model"))
        return self._default_context

    @property
    def default_body_context(self):
        if not self._default_body_context:
            contexts = self.get_entities_by_type("IfcGeometricRepresentationSubContext")
            for context in contexts:
                if context.ContextIdentifier == "Body":
                    self._default_body_context = context
                    return self._default_body_context

            self._default_body_context = self.from_entity(
                run(
                    "context.add_context",
                    self._file,
                    context_type="Model",
                    context_identifier="Body",
                    target_view="MODEL_VIEW",
                    parent=self.default_context.entity,
                )
            )

        return self._default_body_context

    @property
    def default_owner_history(self):
        # We will create a new owner history since we are updating the file
        if not self._default_owner_history:
            person = self._create_entity("IfcPerson")
            organization = self._create_entity("IfcOrganization", Name="compas.dev")
            person_and_org = self._create_entity("IfcPersonAndOrganization", ThePerson=person, TheOrganization=organization)
            application = self._create_entity(
                "IfcApplication",
                ApplicationDeveloper=organization,
                Version=compas_ifc.__version__,
                ApplicationFullName="compas_ifc",
                ApplicationIdentifier="compas_ifc v" + compas_ifc.__version__,
            )

            owner_history = self._create_entity(
                "IfcOwnerHistory",
                OwningUser=person_and_org,
                OwningApplication=application,
                ChangeAction="ADDED",
                CreationDate=int(time.time()),
            )

            self._default_owner_history = owner_history
        return self._default_owner_history



from ifcopenshell import entity_instance


class Attribute:
    def __init__(self, name, type, optional):
        self.name = name
        self.type = type
        self.optional = optional

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        else:
            return self._get_attribute(obj, self.name)

    def __set__(self, obj, value):
        pass

    def _get_attribute(self, obj, name):
        attr = getattr(obj.entity, name)
        if isinstance(attr, (list, tuple)):
            return [self._get_attribute_value(obj, item) for item in attr]
        else:
            return self._get_attribute_value(obj, attr)

    def _get_attribute_value(self, obj, attr):
        if isinstance(attr, entity_instance):
            return obj.reader.from_entity(attr)
        else:
            return attr



from typing import TYPE_CHECKING
from typing import Generator

from compas.data import Data
from compas.geometry import Transformation
from compas.tolerance import TOL

from compas_ifc.file import IFCFile

if TYPE_CHECKING:
    from compas_ifc.entities.base import Base
    from compas_ifc.entities.generated.IFC4 import IfcBuilding
    from compas_ifc.entities.generated.IFC4 import IfcBuildingElement
    from compas_ifc.entities.generated.IFC4 import IfcBuildingStorey
    from compas_ifc.entities.generated.IFC4 import IfcProject
    from compas_ifc.entities.generated.IFC4 import IfcSite


class Model(Data):
    def __init__(self, filepath=None, schema=None, use_occ=False, load_geometries=True, verbose=True):
        self.file = IFCFile(self, filepath=filepath, schema=schema, use_occ=use_occ, load_geometries=load_geometries, verbose=verbose)
        if filepath:
            self.update_linear_deflection()

    @property
    def schema(self):
        return self.file.schema

    @property
    def schema_name(self):
        return self.file.schema_name

    @property
    def entities(self) -> Generator["Base", None, None]:
        for entity in self.file._file:
            yield self.file.from_entity(entity)

    @property
    def project(self) -> "IfcProject":
        return self.file.get_entities_by_type("IfcProject")[0]

    @property
    def sites(self) -> list["IfcSite"]:
        return self.file.get_entities_by_type("IfcSite")

    @property
    def buildings(self) -> list["IfcBuilding"]:
        return self.file.get_entities_by_type("IfcBuilding")

    @property
    def building_storeys(self) -> list["IfcBuildingStorey"]:
        return self.file.get_entities_by_type("IfcBuildingStorey")

    @property
    def building_elements(self) -> list["IfcBuildingElement"]:
        return self.file.get_entities_by_type("IfcBuildingElement")

    @property
    def unit(self):
        length_unit = self.project.length_unit
        if length_unit.Name == "METRE" and length_unit.Prefix == "MILLI":
            return "mm"
        elif length_unit.Name == "METRE" and length_unit.Prefix == "CENTI":
            return "cm"
        elif length_unit.Name == "METRE" and not length_unit.Prefix:
            return "m"

    @unit.setter
    def unit(self, value):
        if value == "mm":
            self.project.length_unit.Prefix = "MILLI"
        elif value == "cm":
            self.project.length_unit.Prefix = "CENTI"
        elif value == "m":
            self.project.length_unit.Prefix = None
        else:
            raise ValueError("Invalid unit. Use 'mm', 'cm', or 'm'.")

    def get_entities_by_type(self, type_name) -> list["Base"]:
        return self.file.get_entities_by_type(type_name)

    def get_entities_by_name(self, name) -> list["Base"]:
        return self.file.get_entities_by_name(name)

    def get_entity_by_global_id(self, global_id) -> "Base":
        return self.file.get_entity_by_global_id(global_id)

    def get_entity_by_id(self, id) -> "Base":
        return self.file.get_entity_by_id(id)

    def print_summary(self):
        """
        Overview of IFC file

        Number of entities
        Size of file
        Project info
        number of key type entities
        Plot of spatial hierarchy"""

        print("=" * 80)
        print("File: {}".format(self.file.filepath))
        print("Size: {} MB".format(self.file.file_size()))
        print("Project: {}".format(self.project.Name))
        print("Description: {}".format(self.project.Description))
        print("Number of sites: {}".format(len(self.sites)))
        print("Number of buildings: {}".format(len(self.buildings)))
        print("Number of building elements: {}".format(len(self.building_elements)))
        print("=" * 80)

    def print_spatial_hierarchy(self, max_depth=3):
        self.project.print_spatial_hierarchy(max_depth=max_depth)

    def save(self, path: str):
        self.file.save(path)

    def export(self, path: str, entities: list["Base"] = [], as_snippet: bool = False, export_materials: bool = True, export_properties: bool = True, export_styles: bool = True):
        self.file.export(path, entities=entities, as_snippet=as_snippet, export_materials=export_materials, export_properties=export_properties, export_styles=export_styles)

    def show(self, entity=None):
        try:
            from compas_viewer import Viewer
            from compas_viewer.components import Treeform

        except ImportError:
            raise ImportError("The show method requires compas_viewer to be installed.")

        viewer = Viewer()
        print(f"Unit: {self.unit}")
        viewer.unit = self.unit

        entity_map = {}

        def parse_entity(entity, parent=None):
            obj = None
            name = f"[{entity.__class__.__name__}]{entity.Name}"
            transformation = Transformation.from_frame(entity.frame) if entity.frame else None
            if getattr(entity, "geometry", None) and not entity.is_a("IfcSpace"):
                obj = viewer.scene.add(entity.geometry, name=name, parent=parent, **entity.style)
                obj.transformation = transformation
            else:
                obj = viewer.scene.add([], name=name, parent=parent)

            obj.attributes["entity"] = entity

            for child in entity.children:
                parse_entity(child, parent=obj)

            if obj:
                entity_map[id(obj)] = entity

        parse_entity(entity or self.project)

        treeform = Treeform()
        viewer.ui.sidebar.widget.addWidget(treeform)

        def update_treeform(form, node):
            entity = node.attributes["entity"]
            treeform.update_from_dict({"Attributes": entity.attributes, "Properties": getattr(entity, "properties", {})})

        viewer.ui.sidebar.sceneform.callback = update_treeform

        viewer.show()

    def create(self, cls="IfcBuildingElementProxy", parent=None, geometry=None, frame=None, properties=None, **kwargs):
        return self.file.create(cls=cls, parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def remove(self, entity):
        self.file.remove(entity)

    def update_linear_deflection(self):
        # TODO: deal with conversion based units like "FOOT"
        if self.unit == "mm":
            TOL.lineardeflection = 1
        elif self.unit == "cm":
            TOL.lineardeflection = 1e-1
        elif self.unit == "m":
            TOL.lineardeflection = 1e-3

    @classmethod
    def template(cls, schema="IFC4", building_count=1, storey_count=1, unit="mm"):
        model = cls(schema=schema)
        project = model.file.default_project
        site = model.create("IfcSite", parent=project, Name="Default Site")
        for i in range(building_count):
            building = model.create("IfcBuilding", parent=site, Name=f"Default Building {i+1}")
            for j in range(storey_count):
                model.create("IfcBuildingStorey", parent=building, Name=f"Default Storey {j+1}")

        model.unit = unit
        model.update_linear_deflection()
        return model


if __name__ == "__main__":
    model = Model("data/wall-with-opening-and-window.ifc")
    model.summary()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from compas_ifc.entities.generated import IfcElement
else:
    IfcElement = object


class IfcElement(IfcElement):
    @property
    def parent(self):
        relations = self.ContainedInStructure()
        if relations:
            return relations[0].RelatingStructure
        else:
            return super().parent



"""
********************************************************************************
representation
********************************************************************************

.. currentmodule:: compas_ifc.representation

Other Functions
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    entity_annotation_geometry
    entity_axis_geometry
    entity_body_geometry
    entity_box_geometry
    entity_clearance_geometry
    entity_lighting_geometry
    entity_mapped_geometry
    entity_reference_geometry
    entity_profile_geometry
    entity_surface_geometry
    entity_survey_geometry


Other Functions
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    tessellation_to_brep
    tessellation_to_mesh

"""

from functools import reduce
from operator import mul
from typing import List

from compas.geometry import Box
from compas.geometry import Scale
from compas.geometry import Transformation

from compas_ifc.brep import TessellatedBrep
from compas_ifc.entities.entity import Entity
from compas_ifc.resources import IfcAxis2Placement3D_to_frame
from compas_ifc.resources import IfcBoundingBox_to_box
from compas_ifc.resources import IfcCartesianTransformationOperator3D_to_frame
from compas_ifc.resources import IfcIndexedPolyCurve_to_lines
from compas_ifc.resources import IfcLocalPlacement_to_frame
from compas_ifc.resources import IfcLocalPlacement_to_transformation
from compas_ifc.resources import IfcShape_to_brep
from compas_ifc.resources import IfcShape_to_tessellatedbrep


# this does not belong here
# but rather in the representations module
def IfcMappedItem_to_transformation(item) -> Transformation:
    """
    Convert an IFC MappedItem to a COMPAS transformation.

    """
    stack = []

    source_frame = IfcAxis2Placement3D_to_frame(item.MappingSource.MappingOrigin)
    target_frame = IfcCartesianTransformationOperator3D_to_frame(item.MappingTarget)

    stack.append(source_frame)
    stack.append(target_frame)

    matrices = [Transformation.from_frame(f) for f in stack]
    return reduce(mul, matrices[::-1])


def entity_transformation(entity: Entity):
    """
    Construct the transformation of an entity.
    """
    return _entity_transformation(entity._entity, scale=entity.model.project.length_scale)


def _entity_transformation(_entity, scale=1.0):
    if _entity.ObjectPlacement:
        scaled_placement = IfcLocalPlacement_to_transformation(_entity.ObjectPlacement, scale=scale)
    else:
        scaled_placement = Scale.from_factors([scale, scale, scale])

    return scaled_placement


def entity_frame(entity: Entity):
    """
    Construct the frame of an entity.
    """
    return IfcLocalPlacement_to_frame(entity._entity.ObjectPlacement)


def entity_body_geometry(entity: Entity, context="Model", use_occ=False, apply_transformation=True):
    """
    Construct the body geometry representations of an entity.

    References
    ----------
    .. [1] :ifc:`body-geometry`

    """
    bodies = []

    representation = None
    if getattr(entity._entity, "Representation", None):
        for temp in entity._entity.Representation.Representations:
            if temp.RepresentationIdentifier == "Body":
                representation = temp
                break

    if not representation:
        return bodies

    if representation.ContextOfItems.ContextType == context:
        for item in representation.Items:
            if use_occ:
                brep = IfcShape_to_brep(item)
                bodies.append(brep)
            else:
                tessellatedbrep = IfcShape_to_tessellatedbrep(item)
                bodies.append(tessellatedbrep)

    if apply_transformation:
        transformation = entity_transformation(entity)
        for body in bodies:
            body.transform(transformation)

    return bodies


def entity_opening_geometry(entity: Entity, use_occ=False, apply_transformation=True):
    """
    Construct the opening geometry representations of an entity.
    """

    voids = []
    if hasattr(entity._entity, "HasOpenings"):
        for opening in entity._entity.HasOpenings:
            element = opening.RelatedOpeningElement

            if apply_transformation:
                transformation = _entity_transformation(element, entity.model.project.length_scale)

            for representation in element.Representation.Representations:
                for item in representation.Items:
                    if use_occ:
                        brep = IfcShape_to_brep(item)
                        if apply_transformation:
                            brep.transform(transformation)
                        voids.append(brep)
                    else:
                        tessellatedbrep = IfcShape_to_tessellatedbrep(item)
                        if apply_transformation:
                            tessellatedbrep.transform(transformation)
                        voids.append(tessellatedbrep)

    return voids


def entity_body_with_opening_geometry(entity: Entity = None, bodies=None, voids=None, context="Model", use_occ=False):
    """
    Construct the body geometry representations of an entity.

    References
    ----------
    .. [1] :ifc:`body-geometry`

    """
    bodies = bodies or entity_body_geometry(entity, context=context, use_occ=use_occ, apply_transformation=True)
    voids = voids or entity_opening_geometry(entity, use_occ=use_occ, apply_transformation=True)

    if not voids:
        return bodies

    if use_occ:
        from compas_occ.brep import OCCBrep
        from OCC.Core.BRep import BRep_Builder
        from OCC.Core.TopoDS import TopoDS_Compound

        print("Using OCC for boolean operations.")
        compound = TopoDS_Compound()
        builder = BRep_Builder()
        builder.MakeCompound(compound)
        for brep in voids:
            builder.Add(compound, brep.occ_shape)
        B = OCCBrep.from_shape(compound)

        shapes = []
        for A in bodies:
            try:
                C: OCCBrep = A - B
                C.make_solid()
                shapes.append(C)
            except Exception:
                shapes.append(A)
                print("Warning: Failed to subtract voids from body.")
    else:
        print("Using CGAL for boolean operations.")
        from compas_cgal.booleans import boolean_difference

        shapes = []
        for A in bodies:
            C = A
            for B in voids:
                vertices, faces = boolean_difference(C.to_vertices_and_faces(), B.to_vertices_and_faces())
                C = TessellatedBrep(vertices=vertices, faces=faces)
            shapes.append(C)

    return shapes


def entity_box_geometry(entity, context="Model") -> List[Box]:
    """
    Construct the box geometry representations of an entity.

    References
    ----------
    .. [1] :ifc:`box-geometry`

    """
    boxes = []

    representation = None

    if getattr(entity._entity, "Representation", None):
        for temp in entity._entity.Representation.Representations:
            if temp.RepresentationIdentifier == "Box":
                representation = temp
                break

    if not representation:
        return boxes

    placement = IfcLocalPlacement_to_transformation(entity._entity.ObjectPlacement)
    scale = entity.model.project.length_scale
    scale = Scale.from_factors([scale, scale, scale])

    if representation.ContextOfItems.ContextType == context:
        if representation.RepresentationType == "BoundingBox":
            for item in representation.Items:
                box = IfcBoundingBox_to_box(item)
                box.transform(placement)
                box.transform(scale)
                boxes.append(box)

    return boxes


def entity_annotation_geometry(entity):
    raise NotImplementedError


def entity_axis_geometry(entity, context="Model"):
    """
    Construct the axis geometry representations of an entity.

    References
    ----------
    .. [1] :ifc:`axis-geometry`

    """
    axes = []
    representation = None

    if getattr(entity._entity, "Representation", None):
        for temp in entity._entity.Representation.Representations:
            if temp.RepresentationIdentifier == "Axis":
                representation = temp
                break

    if not representation:
        return axes

    placement = IfcLocalPlacement_to_transformation(entity._entity.ObjectPlacement)
    scale = entity.model.project.length_scale
    scale = Scale.from_factors([scale, scale, scale])

    if representation.ContextOfItems.ContextType == context:
        if representation.RepresentationType == "Curve2D":
            for item in representation.Items:
                # TODO: deal with other types of curves
                if item.is_a("IfcIndexedPolyCurve"):
                    axis = IfcIndexedPolyCurve_to_lines(item)
                    for line in axis:
                        line.transform(placement)
                        line.transform(scale)
                        axes.append(line)

    return axes


def entity_profile_geometry(entity):
    pass


def entity_surface_geometry(entity):
    pass


def entity_reference_geometry(entity):
    pass


def entity_clearance_geometry(entity):
    pass


def entity_lighting_geometry(entity):
    pass


def entity_survey_geometry(entity):
    pass


def entity_mapped_geometry(entity):
    pass



from typing import List

from compas_ifc.entities.buildingelements import BuildingElement
from compas_ifc.entities.spatialelement import SpatialElement


class Space(SpatialElement):
    """
    Class representing an IFC space.

    Attributes
    ----------
    building_elements : List[:class:`BuildingElement`]
        The building elements contained in this space.

    """

    def __init__(self, entity, model):
        super().__init__(entity, model)
        self._building_elements = None

    @property
    def building_elements(self) -> List[BuildingElement]:
        if not self._building_elements:
            self._building_elements = [entity for entity in self.traverse() if entity.is_a("IfcBuildingElement")]
        return self._building_elements



from typing import Dict
from typing import Union

# from typing import Optional
import ifcopenshell
import ifcopenshell.util.element
import ifcopenshell.util.pset


class Entity:
    """
    Class representing a general IFC entity. Each entity corresponds to a line in IFC file.

    Parameters
    ----------
    entity : :class:`ifcopenshell.entity_instance`
    model : :class:`compas_ifc.model.Model`

    Attributes
    ----------
    ifc_type : str
        The IFC type of the entity.
    declaration : :class:`ifcopenshell.ifcopenshell_wrapper.schema.declaration`
        The IFC declaration of the entity.
    model : :class:`compas_ifc.model.Model`
        The model the entity belongs to.
    attributes : Dict[str, Union[str, float, int, list, tuple]]
        All the attributes of the entity.
    psets : Dict[str, Dict[str, Union[str, float, int, list, tuple]]]
        All the property sets of the entity.
    properties : Dict[str, Union[str, float, int, list, tuple]]
        All the properties of all property sets compiled into a single dict.

    """

    def __init__(self, entity: ifcopenshell.entity_instance, model):
        self._entity = entity
        self._model = model
        self._declaration = None
        self._psets = {}
        self._attributes = {}
        self._properties = {}
        self._ifc_type = None

    def __repr__(self):
        return "<{}>".format(self.ifc_type)

    @property
    def declaration(self):
        if not self._declaration and self.model.schema:
            self._declaration = self.model.schema.declaration_by_name(self.ifc_type)
        return self._declaration

    @property
    def id(self):
        if self._entity:
            return self._entity.id()
        return None

    @property
    def ifc_type(self):
        if self._entity:
            self._ifc_type = self._entity.is_a()
        else:
            self._ifc_type = "Ifc" + type(self).__name__
        return self._ifc_type

    def is_a(self, ifc_type: str = None):
        if not ifc_type:
            return self.ifc_type
        if not self._entity:
            declaration = self.declaration
            inheritance = [declaration.name()]
            while hasattr(declaration, "supertype") and declaration.supertype():
                inheritance.append(declaration.supertype().name())
                declaration = declaration.supertype()
            return ifc_type in inheritance
        return self._entity.is_a(ifc_type)

    def __getitem__(self, key: str):
        return self.attribute(key)

    def __setitem__(self, key: str, value):
        self.set_attribute(key, value)

    def _collect_attributes(self):
        if not self._entity:
            return {attr.name(): None for attr in self.declaration.all_attributes()}

        attributes = self._entity.get_info(
            recursive=False,
            include_identifier=False,
        )
        del attributes["type"]
        return attributes

    def _collect_properties(self):
        properties = {}
        for pset in self.psets.values():
            for name in pset:
                properties[name] = pset[name]
        return properties

    def _collect_psets(self):
        psets = {}
        if not self._entity:
            return psets
        _psets = ifcopenshell.util.element.get_psets(self._entity)
        for name in _psets:
            psets[name] = _psets[name]
        return psets

    @property
    def model(self):
        return self._model

    @property
    def attributes(self) -> Dict:
        if not self._attributes:
            self._attributes = self._collect_attributes()

        # TODO: this is a bit of a hack, need to clean up
        for key, value in self._attributes.items():
            if isinstance(value, ifcopenshell.entity_instance):
                self._attributes[key] = self.model.reader.get_entity(value)
            elif isinstance(value, tuple) and value and isinstance(list(value)[0], ifcopenshell.entity_instance):
                self._attributes[key] = [self.model.reader.get_entity(v) for v in value]
        return self._attributes

    @property
    def psets(self) -> Dict:
        if not self._psets:
            self._psets = self._collect_psets()
        return self._psets

    @psets.setter
    def psets(self, psets: Dict):
        self._psets = psets

    @property
    def properties(self) -> Dict:
        if not self._properties:
            self._properties = self._collect_properties()
        return self._properties

    def has_attribute(self, name: str) -> bool:
        """
        Verify that the entity has a specific attributes.

        Returns
        -------
        bool

        """
        return name in self.attributes

    def attribute(self, name: str) -> Union[str, int, float]:
        """
        Get the value of a named attribute.

        Parameters
        ----------
        name : str
            The name of the attribute.

        Returns
        -------
        str | int | float
            The value of the attribute.

        """
        return self.attributes.get(name)

    def set_attribute(self, name: str, value: Union[str, int, float]) -> None:
        """
        Set the value of a named attribute.

        Parameters
        ----------
        name : str
            The name of the attribute.
        value : str | int | float
            The value of the attribute.

        Returns
        -------
        None

        """
        self.attributes[name] = value

    def set_attributes(self, attributes: Dict[str, Union[str, int, float]]) -> None:
        """
        Set the values of multiple attributes.

        Parameters
        ----------
        attributes : Dict[str, Union[str, int, float]]
            The attributes to set.

        Returns
        -------
        None

        """
        for name, value in attributes.items():
            self.set_attribute(name, value)

    def pset(self, name: str) -> Dict:
        """
        Get the property set with the given name.

        Parameters
        ----------
        name : str

        Returns
        -------
        Dict

        """
        return self._psets[name]

    def has_property(self, name: str) -> bool:
        """
        Verify that this entity has a specific property.

        Parameters
        ----------
        name : str

        Returns
        -------
        bool

        """
        return name in self.properties

    def property(self, name: str) -> Union[str, int, float]:
        """
        Get the value of the property with the given name.

        Parameters
        ----------
        name : str

        Returns
        -------
        str | int | float

        """
        return self.properties.get(name)

    def set_property(self, name: str, value: Union[str, int, float]) -> None:
        """
        Set the value of the property with a given name.

        Parameters
        ----------
        name : str
            The name of the property.
        value : str | int | float
            The value of the property.

        Returns
        -------
        None

        """
        self.properties[name] = value

    def inheritance(self):
        """
        Find the ancestors of the current entity up to the root element.

        Returns
        -------
        List[str]

        """
        declaration = self.declaration
        inheritance = [declaration.name()]
        while declaration.supertype():
            inheritance.append(declaration.supertype().name())
            declaration = declaration.supertype()
        return inheritance[::-1]

    def print_inheritance(self) -> None:
        """
        Print the entity inheritance as a nested list.

        Returns
        -------
        None

        """
        for index, item in enumerate(self.inheritance()):
            print("-" * (index + 1) + f" {item}")

    def traverse_branch(self):
        yield self



"""
********************************************************************************
helpers
********************************************************************************

.. currentmodule:: compas_ifc.helpers

Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    public_attributes

"""

import ifcopenshell


def public_attributes(o):
    public = [name for name in dir(o) if not name.startswith("_")]
    return public


def print_schema_ref_tree(schema, declaration, indent=0, only_required=True):
    """Prints a tree view of the schema references."""

    if indent == 0:
        print(declaration)

    indent += 4

    if declaration.as_entity():
        for attribute in declaration.all_attributes():
            if only_required and attribute.optional():
                continue

            if attribute.type_of_attribute().as_named_type():
                print(" " * indent, attribute)
                print_schema_ref_tree(schema, attribute.type_of_attribute().declared_type(), indent)
            elif attribute.type_of_attribute().as_aggregation_type():
                print(" " * indent, attribute)
            else:
                raise NotImplementedError


def print_entity_ref_tree(entity, indent=0):
    """Prints a tree view of the entity references."""
    info = entity.get_info()
    print(" " * indent, "<", entity.id(), entity.is_a(), ">")
    for key, value in info.items():
        if key.lower() == "type" or key.lower() == "id":
            continue
        if isinstance(value, ifcopenshell.entity_instance):
            print(" " * indent, key, ":")
            print_entity_ref_tree(value, indent + 4)
        else:
            print(" " * indent, key, ":", value)



import numpy as np
from compas.geometry import Geometry
from compas.geometry import transform_points_numpy


class TessellatedBrep(Geometry):
    def __init__(self, vertices=None, edges=None, faces=None, **kwargs):
        super().__init__(**kwargs)
        if vertices is None:
            vertices = []
        if edges is None:
            edges = []
        if faces is None:
            faces = []
        self.vertices = np.array(vertices).reshape(-1, 3)
        self.edges = np.array(edges).reshape(-1, 2)
        self.faces = np.array(faces).reshape(-1, 3)

    def transform(self, transformation):
        self.vertices = transform_points_numpy(self.vertices, transformation)

    def to_vertices_and_faces(self):
        return self.vertices, self.faces
